#include <cuda.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

using namespace std;
 

__global__ void allocateMatrix(float* matrix, size_t pitch, int width, int height, float argument)
{
	//This kernel needs to be called with WIDTH blocks and HEIGHT threads
	if(blockIdx.x < height && threadIdx.x < width) 
	{ 
		if(blockIdx.x == threadIdx.x)
		{
			float* element = (float*)((char*)matrix + threadIdx.x*pitch) + blockIdx.x;
			*element = 2;//argument + threadIdx.x;//row[threadIdx.x] = argument;
		}
	}
}

__global__ void squareDiagonalMatrix(float* matrix, size_t pitch, int size)
{
	if(blockIdx.x < size && threadIdx.x < size)
	{
		if(blockIdx.x == threadIdx.x)
		{
			float* element = (float*)((char*)matrix + threadIdx.x*pitch) + blockIdx.x;
			*element = (*element)*(*element);//argument + threadIdx.x;//row[threadIdx.x] = argument;
		}
	}
}
__global__ void squareMatrix(float* matrix1, float* matrix2, size_t pitch, int size)
{
	//Run with SIZE blocks and SIZE threads
	//Assumes matrix1 is a square matrix
	//Therefore, width = height = size

	if(threadIdx.x < size && blockIdx.x < size)
	{
		float* row = (float*)((char*)matrix2+blockIdx.x*pitch); //"Prime" new matrix
		row[threadIdx.x] = 0; 

		//X=THREAD
		//Y=BLOCK

		for(int z = 0; z < size; z++)
		{
			float* m1 = (float*)((char*)matrix1+threadIdx.x*pitch);
			float M1 = m1[z];
			float* m2 = (float*)((char*)matrix1+z*pitch);
			float M2 = m2[blockIdx.x];
			row[threadIdx.x]+= (M1*M2);
		}
	}
}

__global__ void matrixMultiply(float* matrix1, float* matrix2, float* matrix3, size_t pitch, int size)
{
	//Assumes matrices are of same size and are square
	//Multiply matrix1 by matrix2 and return matrix3
	if(threadIdx.x < size && blockIdx.x < size)
	{
		float* row = (float*)((char*)matrix3+blockIdx.x*pitch);
		row[threadIdx.x] = 0;

		for(int z = 0; z < size; z++)
		{
			float* m1 = (float*)((char*)matrix1+threadIdx.x*pitch);
			float M1 = m1[z];
			float* m2 = (float*)((char*)matrix2+z*pitch);
			float M2 = m2[blockIdx.x];
			row[threadIdx.x]+=(M1*M2);
		}
	}
}

int main()
{
	const int width = 10;
	const int height = 10;
	const int depth = 10;

	float* matrix;
	float* anotherMatrix;
	float* aThirdMatrix;
	float* squaredMatrix;
	size_t pitch, pitch2, pitch3,pitch4; /*size of 1 row - DUMMY VARIABLE*/
	float testArray[width][height];
	for(int x = 0; x < width; x++)
		for(int y = 0; y < height; y++)
		{
			testArray[x][y] = 0;
		}


	cudaMallocPitch((void**)&matrix, &pitch, width*sizeof(float), height);
	cudaMallocPitch((void**)&anotherMatrix, &pitch2, width*sizeof(float), height);
	cudaMallocPitch((void**)&aThirdMatrix, &pitch3, width*sizeof(float), height);
	cudaMallocPitch((void**)&squaredMatrix, &pitch4, width*sizeof(float), height);

	float test = 1;
	allocateMatrix<<<width,height>>>(matrix, pitch, width, height, test); //Create matrix
	allocateMatrix<<<width,height>>>(anotherMatrix, pitch2, width, height,3); //Create matrix
	
	cudaMemcpy2D((void*)testArray, width*sizeof(float), (void*)matrix, pitch, width*sizeof(float), height, cudaMemcpyDeviceToHost);
	cout << "Matrix 1:" << endl;
	for(int x = 0; x < width; x++)
	{
		for(int y = 0; y < height; y++)
		{
			printf("%.0f ", testArray[x][y]);
		}
		printf("\n");
	}
	cout << endl;	
	cudaMemcpy2D((void*)testArray, width*sizeof(float), (void*)anotherMatrix, pitch, width*sizeof(float), height, cudaMemcpyDeviceToHost);
	cout << "Matrix 2:" << endl;
	for(int x = 0; x < width; x++)
	{
		for(int y = 0; y < height; y++)
		{
			printf("%.0f ", testArray[x][y]);
		}
		printf("\n");
	}
	cout << endl;

	//matrixMultiply<<<width,width>>>(matrix,anotherMatrix,aThirdMatrix,pitch, width);
	squareDiagonalMatrix<<<width,width>>>(matrix, pitch,width);
	cudaMemcpy2D((void*)testArray, width*sizeof(float), (void*)matrix, pitch, width*sizeof(float), height, cudaMemcpyDeviceToHost);
	
	cout << "Squared Matrix:" << endl;
	for(int x = 0; x < width; x++)
	{
		for(int y = 0; y < height; y++)
		{
			printf("%.0f ", testArray[x][y]);
		}
		printf("\n");
	}

	//free(input);
	system("PAUSE");
	cudaFree(matrix);
	return 0;
}